#include <kos.h>
#include <stdio.h>
#include <assert.h>
#include <math.h>
#include "libsh4.h"
#include "profile.h"
#include "watchdog.h"
#include "npvr_vertex.h"
#include "pvr_cxt.h"
#include "xmtrx.h"

#define DBG(a...)	do { printf(a); fflush(stdout); } while(0)

#define MB	(1024*1024)
#define MEGA	(1000*1000)

//This should be a power of 2
char __attribute__ ((aligned (32))) rambuff[8*MB];

#define SQ_UNITS (sizeof(rambuff) / 32)

void pvr_setup() {
	static pvr_init_params_t pvr_params = {
		{ PVR_BINSIZE_32, PVR_BINSIZE_0, PVR_BINSIZE_0, PVR_BINSIZE_0, PVR_BINSIZE_0 },
		2.6 * MB,
		0,	//dma
		0,	//fsaa
		0	//autosort disable
	};
	
	pvr_init (&pvr_params);
	
	//PVR needs to warm up for a frame, or results will be low
	pvr_wait_ready();
	pvr_scene_begin();
	pvr_scene_finish();
	
	//Prepare PVR for benchmark
	pvr_wait_ready();
	pvr_scene_begin();
	pvr_list_begin(PVR_LIST_OP_POLY);
}

void pvr_setup_tex() {
	static pvr_init_params_t pvr_params = {
		{ PVR_BINSIZE_32, PVR_BINSIZE_0, PVR_BINSIZE_0, PVR_BINSIZE_0, PVR_BINSIZE_0 },
		0.25 * MB,
		0,	//dma
		0,	//fsaa
		0	//autosort disable
	};
	pvr_init (&pvr_params);
	pvr_wait_ready();
	pvr_scene_begin();
	pvr_list_begin(PVR_LIST_OP_POLY);
}

void sq_main_ram_write_test() {
	DBG("---RAM SQ Bandwidth Test---\n");
	
	//Run benchmark
	size_t i = SQ_UNITS;
	int *sq = sqPrepare(rambuff);
	pfMicroseconds ramsqtime = pfStart();
	do {
		sq = sqSubmit32Advance(sq);
	} while(--i);
	ramsqtime = pfEnd(ramsqtime);
	
	float ramtime_sec = pfSecs(ramsqtime);
	DBG("RAM SQ Bandwidth: %.2f Mbyte/sec\n", sizeof(rambuff) / ramtime_sec / MEGA);
}

void cache_main_ram_write_test() {
	DBG("---RAM Cache Bandwidth Test---\n");
	
	//Run benchmark
	size_t i = SQ_UNITS;
	char *dst = (char*)rambuff;
	pfMicroseconds ramcachetime = pfStart();
	do {
		sh4CachelineAllocate(dst, 0);
		sh4Writeback(dst);
		dst += 32;
	} while(--i);
	ramcachetime = pfEnd(ramcachetime);

	float ramcachetime_sec = pfSecs(ramcachetime);
	DBG("RAM cache Bandwidth: %.2f Mbyte/sec\n", sizeof(rambuff) / ramcachetime_sec / MEGA);
}

//Adjust to test bandwidth of different strip lengths
//Set to a real high number to test endless strips
//Must be power of 2
#define EOS_EVERY_X_VERTS	(1<<24)
//#define EOS_EVERY_X_VERTS	(4)

//No need to adjust these
#define EOS_MASK	(EOS_EVERY_X_VERTS-1)
#define TRI_PER_VERT	((float)EOS_EVERY_X_VERTS / (float)(EOS_EVERY_X_VERTS - 2))

void set_cxt_opaque_poly() {
	pvr_context cxt;
	pc_set_default_polygon(&cxt);
	pc_set_color_format(&cxt, PC_PACKED);
	pc_set_list(&cxt, PC_OPAQUE_POLY);
	sqCopy32(sqPrepare((void*)PVR_TA_INPUT), &cxt);
}

void ta_off_screen_sq_test() {
	DBG("---TA Off Screen SQ Test---\n");
	
	pvr_setup();
	set_cxt_opaque_poly();

	//Initialize both SQs to dummyoff-screen vertex
	pvr_vertex32 *dummyvert = sqPrepare((void*)PVR_TA_INPUT);
	pv_set_cmd_vertex(dummyvert);
	pv_set_pos(dummyvert, -100, -100, 0.5);
	pv_set_cmd_vertex_eos(dummyvert+1);
	pv_set_pos(dummyvert+1, -100, -100, 0.5);
	
	//Run benchmark
	size_t i = SQ_UNITS;
	pfMicroseconds pvrsqtime = pfStart();
	do {
		if ((i & EOS_MASK) == 0)
			sqSubmit32(dummyvert+1);
		else
			sqSubmit32(dummyvert);
	} while(--i);
	pvrsqtime = pfEnd(pvrsqtime);

	float pvrsqtime_sec = pfSecs(pvrsqtime);
	DBG("PVR TA SQ Off Screen Bandwidth: %.2f Mbyte/sec\n", sizeof(rambuff) / pvrsqtime_sec / MEGA);
	DBG("PVR TA SQ Off Screen Vert Speed: %.2f MVert/sec\n", SQ_UNITS / pvrsqtime_sec / MEGA);
	DBG("PVR TA SQ Off Screen Tri Speed: %.2f MTri/sec\n", SQ_UNITS / TRI_PER_VERT / pvrsqtime_sec / MEGA);

	pvr_shutdown();
}

void ta_off_screen_dma_test() {
	DBG("---TA Off Screen DMA Test---\n");
	
	pvr_setup();
	set_cxt_opaque_poly();
	
	//Initialize DMA source
	size_t i;
	pvr_vertex32 *v = (pvr_vertex32*)rambuff;
	for(i = 0; i < SQ_UNITS; i++) {
		pv_set_cmd_vertex(v);
		if ((i & EOS_MASK) == EOS_MASK)
			pv_set_cmd_vertex_eos(v);
		pv_set_pos(v, -100, -100, 0.5);
		sh4Writeback(v);
		v++;
	}
	
	//Run benchmark
	pfMicroseconds pvrdmatime = pfStart();
	pvr_dma_load_ta(rambuff, sizeof(rambuff), 1, NULL, 0);
	pvrdmatime = pfEnd(pvrdmatime);
	
	float pvrdmatime_sec = pfSecs(pvrdmatime);
	DBG("PVR TA DMA Off Screen Bandwidth: %.2f Mbyte/sec\n", sizeof(rambuff) / pvrdmatime_sec / MEGA);
	DBG("PVR TA DMA Off Screen Vert Speed: %.2f Mvert/sec\n", SQ_UNITS / pvrdmatime_sec / MEGA);
	DBG("PVR TA DMA Off Screen Tri Speed: %.2f MTri/sec\n", SQ_UNITS / TRI_PER_VERT / pvrdmatime_sec / MEGA);
	
	pvr_shutdown();
}

#define ON_SCREEN_VERTS	(16000)
void ta_on_screen_sq_test() {
	DBG("---TA On Screen SQ Test---\n");
		
	const int verts = ON_SCREEN_VERTS;
	
	pvr_setup();
	set_cxt_opaque_poly();

	//Initialize both SQs to dummyoff-screen vertex
	pvr_vertex32 *dummyvert = sqPrepare((void*)PVR_TA_INPUT);
	pv_set_cmd_vertex(dummyvert);
	pv_set_pos(dummyvert, 100, 100, 0.5);
	pv_set_cmd_vertex_eos(dummyvert+1);
	pv_set_pos(dummyvert+1, 100, 100, 0.5);

	//Run benchmark
	size_t i = verts;
	pfMicroseconds pvrsqtime = pfStart();
	do {
		if ((i & EOS_MASK) == 0)
			sqSubmit32(dummyvert+1);
		else
			sqSubmit32(dummyvert);
	} while(--i);
	pvrsqtime = pfEnd(pvrsqtime);

	float pvrsqtime_sec = pfSecs(pvrsqtime);
	DBG("PVR TA SQ On Screen Bandwidth: %.2f Mbyte/sec\n", verts * 32 / pvrsqtime_sec / MEGA);
	DBG("PVR TA SQ On Screen Vert Speed: %.2f MVert/sec\n", verts / pvrsqtime_sec / MEGA);
	DBG("PVR TA SQ On Screen Tri Speed: %.2f MTri/sec\n", verts / TRI_PER_VERT / pvrsqtime_sec / MEGA);

	pvr_shutdown();
}

void ta_on_screen_dma_test() {
	DBG("---TA On Screen DMA Test---\n");
	
	const int verts = ON_SCREEN_VERTS;
	
	pvr_setup();
	set_cxt_opaque_poly();
	
	//Initialize DMA source
	size_t i;
	pvr_vertex32 *v = (pvr_vertex32*)rambuff;
	for(i = 0; i < verts; i++) {
		pv_set_cmd_vertex(v);
		if ((i & EOS_MASK) == EOS_MASK)
			pv_set_cmd_vertex_eos(v);
		pv_set_pos(v, 100, 100, 0.5);
		sh4Writeback(v);
		v++;
	}
	
	//Run benchmark
	pfMicroseconds pvrdmatime = pfStart();
	pvr_dma_load_ta(rambuff, verts * 32, 1, NULL, 0);
	pvrdmatime = pfEnd(pvrdmatime);
	
	float pvrdmatime_sec = pfSecs(pvrdmatime);
	DBG("PVR TA DMA On Screen Bandwidth: %.2f Mbyte/sec\n", verts * 32 / pvrdmatime_sec / MEGA);
	DBG("PVR TA DMA On Screen Vert Speed: %.2f Mvert/sec\n", verts / pvrdmatime_sec / MEGA);
	DBG("PVR TA DMA On Screen Tri Speed: %.2f MTri/sec\n", verts / TRI_PER_VERT / pvrdmatime_sec / MEGA);
	
	pvr_shutdown();
}

#define BIG_POLY_CNT	1500
void ta_big_poly_sq_test() {
	DBG("---TA Big Poly SQ Test---\n");
	
	pvr_setup();
	set_cxt_opaque_poly();
	
	//Run benchmark
	pvr_vertex32 *v = sqPrepare((void*)PVR_TA_INPUT);
	pfMicroseconds pvrbigpolytime = pfStart();
	int i, verts = BIG_POLY_CNT;
	for(i = 0; i < verts/4; i++) {
		pv_set_pos(v, 0, 0, 0.5);
		pv_set_cmd_submit_vertex(v);
		v++;
		
		pv_set_pos(v, 0, 480, 0.5);
		pv_set_cmd_submit_vertex(v);
		v++;
		
		pv_set_pos(v, 640, 0, 0.5);
		pv_set_cmd_submit_vertex(v);
		v++;
		
		pv_set_pos(v, 640, 480, 0.5);
		pv_set_cmd_submit_vertex_eos(v);
		v++;
	}
	pvrbigpolytime = pfEnd(pvrbigpolytime);
	
	float pvrbigpolytime_sec = pfSecs(pvrbigpolytime);
	DBG("PVR SQ Big Poly Bandwidth: %.2f Mbyte/sec\n", (verts*32) / pvrbigpolytime_sec / MEGA);
	pvr_shutdown();
}

void ta_big_poly_dma_test() {
	DBG("---TA Big Poly DMA Test---\n");
	
	//Initialize DMA source
	int i,  verts = BIG_POLY_CNT;
	pvr_vertex32 *v = (pvr_vertex32*)rambuff;
	for(i = 0; i < verts/4; i++) {
		pv_set_pos(v, 0, 0, 0.5);
		pv_set_cmd_vertex(v);
		sh4Writeback(v);
		v++;
		
		pv_set_pos(v, 0, 480, 0.5);
		pv_set_cmd_vertex(v);
		sh4Writeback(v);
		v++;
		
		pv_set_pos(v, 640, 0, 0.5);
		pv_set_cmd_vertex(v);
		sh4Writeback(v);
		v++;
		
		pv_set_pos(v, 640, 480, 0.5);
		pv_set_cmd_submit_vertex_eos(v);
		sh4Writeback(v);
		v++;
	}
	
	pvr_setup();
	set_cxt_opaque_poly();
	
	//Run benchmark
	pfMicroseconds pvrbigpolytime = pfStart();
	pvr_dma_load_ta(rambuff, verts * 32, 1, NULL, 0);
	pvrbigpolytime = pfEnd(pvrbigpolytime);
	
	float pvrbigpolytime_sec = pfSecs(pvrbigpolytime);
	DBG("PVR DMA Big Poly Bandwidth: %.2f Mbyte/sec\n", (verts*32) / pvrbigpolytime_sec / MEGA);
	pvr_shutdown();
}

void set_cxt_opaque_spr() {
	pvr_context cxt;
	pc_set_default_sprite(&cxt);
	pc_set_list(&cxt, PC_OPAQUE_POLY);
	sqCopy32(sqPrepare((void*)PVR_TA_INPUT), &cxt);
}
#define SPRITE_VERT_SIZE	(sizeof(pvr_vertex64s))
void ta_big_sprite_sq_test() {
	DBG("---TA Big Sprite SQ Test---\n");
	pvr_setup();
	
	void *pvrsq = sqPrepare((void*)PVR_TA_INPUT);
	
	set_cxt_opaque_spr();
	
	pvr_vertex64s *v = (pvr_vertex64s*)rambuff;
	pv_set_cmd_vertex_eos(v);
	pv_set_rect(v, 0, 0, 640, 480, .5);
	
	pfMicroseconds pvrbigpolytime = pfStart();
	int i, sprites = BIG_POLY_CNT, spritebytes = sprites * SPRITE_VERT_SIZE;
	for(i = 0; i < sprites; i++) {
		sqSubmit64A(pvrsq);
		sqSubmit64B(pvrsq);
	}
	pvrbigpolytime = pfEnd(pvrbigpolytime);
	
	float pvrbigpolytime_sec = pfSecs(pvrbigpolytime);
	DBG("PVR SQ Big Sprite Bandwidth: %.2f Mbyte/sec\n", spritebytes / pvrbigpolytime_sec / MEGA);
	pvr_shutdown();
}

void ta_big_sprite_dma_test() {
	DBG("---TA Big Sprite DMA Test---\n");
	
	int i,  sprites = BIG_POLY_CNT, spritebytes = sprites * SPRITE_VERT_SIZE;
	pvr_vertex64s *v = (pvr_vertex64s*)rambuff;
	for(i = 0; i < sprites; i++) {
		pv_set_rect(v, 0, 0, 640, 480, .5);
		pv_set_cmd_vertex_eos(v);
		v++;
	}
	dcache_flush_range((uint32)rambuff, spritebytes);
	
	pvr_setup();
	
	set_cxt_opaque_spr();
	
	pfMicroseconds pvrbigpolytime = pfStart();
	pvr_dma_load_ta(rambuff, spritebytes, 1, NULL, 0);
	pvrbigpolytime = pfEnd(pvrbigpolytime);
	
	float pvrbigpolytime_sec = pfSecs(pvrbigpolytime);
	DBG("PVR DMA Big Sprite Bandwidth: %.2f Mbyte/sec\n", spritebytes / pvrbigpolytime_sec / MEGA);
	pvr_shutdown();
}

void texture_dma_test() {
	DBG("---Texture DMA Test---\n");
	pvr_setup_tex();
	
	int dmasize = 4*1024*1024;
	
	pvr_ptr_t *tex = pvr_mem_malloc(dmasize);
	
	pfMicroseconds pvrdmatime = pfStart();
	pvr_txr_load_dma(rambuff, tex, dmasize, 1, NULL, 0);
	pvrdmatime = pfEnd(pvrdmatime);
	
	float pvrdmatime_sec = pfSecs(pvrdmatime);
	DBG("PVR VRAM64 DMA Bandwidth: %.2f Mbyte/sec\n", dmasize / pvrdmatime_sec / MEGA);
	
	pvr_shutdown();
}

#define INITIAL_VALUE	(0x00c0ffee)
#define SQ_VALUE	(0xdeafbeef)
void sq_cache_test() {
	int previrq = irq_disable();
	
	//This should really be done from assembler, so that the compiler can't
	//accidently load from RAM when if it decides to load a constant in the
	//wrong place, but for this case it's probably ok
	
	//Can only access SQ with 4/8 byte access, to cast buffer to int
	volatile unsigned int *vrambuff = (volatile unsigned int *)rambuff;
	volatile unsigned int *sq = sqPrepare(rambuff);
	unsigned int firstread, secondread;
	
	//Initialize cache to 0x00c0ffee
	vrambuff[0] = INITIAL_VALUE;
	
	//Prepare SQ for writing 0xdeadbeef
	sq[0] = SQ_VALUE;
	
	//Submit SQ
	sqSubmit32((unsigned int *)sq);
	
	//Read from cache
	firstread = vrambuff[0];
	
	//Invalidate cache and read from RAM
	sh4Invalidate((void*)vrambuff);
	secondread = vrambuff[0];
	
	DBG("Cache read: %08x\nCache invalidate read: %08x\n", firstread, secondread);
	if (firstread == INITIAL_VALUE) {
		DBG("SQ did NOT flush cache, we got the old value\n");
		if (secondread == SQ_VALUE)
			DBG("We got the new results from the SQ after invalidating the cacheline and reloading\n");
		else
			DBG("We got something completely unexpected the second read\n");
	} else if (firstread == SQ_VALUE) {
		DBG("SQ did flush cache, we got the new value\n");			
	} else {
		DBG("Something weird happened\n");
	}
	
	irq_restore(previrq);
}

struct wide_vertex {
	float x,y,z,w;
	float u;
	float nx,ny,nz,nw;
	float v;
};

struct wide_vertex * toruswide (float TubeRadius, float Radius, int Sides, int Rings )
{
	int i,j,pos = 0;


	float theta, phi, theta1, cosTheta, sinTheta, cosTheta1, sinTheta1, ringDelta, sideDelta, cosPhi, sinPhi, dist;

	sideDelta = 2.0 * M_PI / Sides;
	ringDelta = 2.0 * M_PI / Rings;
	theta = 0.0;
	cosTheta = 1.0;
	sinTheta = 0.0;

	struct wide_vertex * torus = memalign(32, (Rings+1) * (Sides+1) * 2 * sizeof(struct wide_vertex));

	for (i = Rings; i >= 0; i--) {
		theta1 = theta + ringDelta;
		cosTheta1 = fcos(theta1);
		sinTheta1 = fsin(theta1);
		phi = 0.0;
		for (j = Sides; j >= 0; j--) {
			phi = phi + sideDelta;
			cosPhi = fcos(phi);
			sinPhi = fsin(phi);
			dist = Radius + (TubeRadius * cosPhi);

			torus[pos].nx = cosTheta1 * cosPhi;
			torus[pos].ny = -sinTheta1 * cosPhi;
			torus[pos].nz = sinPhi;
			torus[pos].nw = 0.0;

			torus[pos].u = cosTheta1;
			torus[pos].v = sinTheta1;

			torus[pos].x = cosTheta1 * dist;
			torus[pos].y = -sinTheta1 * dist;
			torus[pos].z = TubeRadius * sinPhi;
			torus[pos++].w = 1;

			torus[pos].nx = cosTheta * cosPhi;
			torus[pos].ny = -sinTheta * cosPhi;
			torus[pos].nz = sinPhi;
			torus[pos].nw = 0;

			torus[pos].u = cosTheta;
			torus[pos].v = sinTheta;

			torus[pos].x = cosTheta * dist;
			torus[pos].y = -sinTheta * dist;
			torus[pos].z = TubeRadius * sinPhi;
			torus[pos++].w = 1;
		}
		theta = theta1;
		cosTheta = cosTheta1;
		sinTheta = sinTheta1;
	}

	return torus;
}

int check_start() {
	maple_device_t *cont;
	cont_state_t *state;

	cont = maple_enum_type(0, MAPLE_FUNC_CONTROLLER);

	if(!cont)
		return 0;

	/* Check for start on the controller */
	state = (cont_state_t *)maple_dev_status(cont);

	if(!state) {
		return 0;
	}

	if(state->buttons & CONT_START) {
		return 1;
	}

	return 0;
}

#define TORUS_X	(10)
#define TORUS_Y (10)
#define TORUS_VERTS ((TORUS_X+1) * (TORUS_Y+1) * 2)
#define OBJ_PER_RING	(10)
#define RINGS	(10)
extern int render_strip_wide(void *targ, struct wide_vertex *strips, int vertcount, float lx, float ly, float lz, float lw, float ambient);
void ta_model_test() {
	pvr_setup();
	pvr_set_bg_color(0,0,0.21);
	
	struct wide_vertex * torus = toruswide(0.08, 0.3, TORUS_X, TORUS_Y);
	
	pfMicroseconds time;
	do {
		pvr_context cxt;
		pc_set_default_polygon(&cxt);
		pc_set_color_format(&cxt, PC_INTENSITY);
		pc_set_intensity_color(&cxt, 1, 1, 1, 1);
		pc_set_list(&cxt, PC_OPAQUE_POLY);
		sqCopy32(sqPrepare((void*)PVR_TA_INPUT), &cxt);
		
		float lx = 0, ly = 1, lz = 0;
		
		static float rott = 0;
		rott += 0.008;
		
		xmtrxIdentity();
		mat_perspective(320.0, 240.0, 1.0, 1000.0, 0.1);
		xmtrxTranslate(0,-10,-25);
		xmtrxRotateFX(-1);
		xmtrxScale(3.4,3.4,3.4);
		xmtrxPush();
		void *targ = sqPrepare((void*)PVR_TA_INPUT);
		
		time = pfStart();
		int i, j;
		for(j=1; j<= RINGS; j++){
			for (i=0; i < OBJ_PER_RING; i++) {
				xmtrxTranslate(0.6f * j * fcos(i * 2*M_PI / OBJ_PER_RING), 0.0f, 0.6f * j * fsin(i * 2*M_PI / OBJ_PER_RING));
				xmtrxRotateFY(rott);
				render_strip_wide(targ, torus, TORUS_VERTS, lx,ly,lz,0.0,0.2);
				xmtrxPeek();
			}
		}
		time = pfEnd(time);
		
		xmtrxDrop();
		
		pvr_list_finish();
		pvr_scene_finish();
		
		if (check_start())
			break;
		
		pvr_wait_ready();
		pvr_scene_begin();
		pvr_list_begin(PVR_LIST_OP_POLY);
	} while(1);
	
	free(torus);
	
	float sec = pfSecs(time);
	int modelpolys = TORUS_X * TORUS_Y * 2;
	int verts = TORUS_VERTS * RINGS * OBJ_PER_RING;
	
	DBG("Render SQ Bandwidth: %.2f Mbyte/sec\n", verts * 32 / sec / MEGA);
	DBG("Render SQ Vert Speed: %.2f Mvert/sec\n", verts / sec / MEGA);
	DBG("Render SQ Tri Speed: %.2f MTri/sec\n", (modelpolys * RINGS * OBJ_PER_RING) / sec / MEGA);
	
	DBG("Model size: %i bytes\n", TORUS_VERTS * 32);
	DBG("Submit: %f ms\n", pfMillisecs(time));
	DBG("Verts per frame: %i\n", verts);
}

xMatrix mtrxstack[64];

int main() {
	//Make sure EOS_EVERY_X_VERTS is a power of 2
	assert((EOS_EVERY_X_VERTS & (EOS_EVERY_X_VERTS-1)) == 0);
	
	DBG("----------SQ/Cache test----------\n");
	sq_cache_test();
	
	DBG("\n----------Bandwidth tester----------\n");
	if (1) {
		sq_main_ram_write_test();
		cache_main_ram_write_test();
		DBG("----EOS EVERY %i VERTS----\n", EOS_EVERY_X_VERTS);
		ta_off_screen_sq_test();
		ta_off_screen_dma_test();
		ta_on_screen_sq_test();
		ta_on_screen_dma_test();
		ta_big_poly_sq_test();
		ta_big_poly_dma_test();
		ta_big_sprite_sq_test();
		ta_big_sprite_dma_test();
		texture_dma_test();
	}
	
	if (0) {
		DBG("\n----------Render test----------\n");
		xmtrxSetStackPointer(mtrxstack, sizeof(mtrxstack));
		ta_model_test();
	}
	
	return 0;
}
